{
  To add a "gasp" record to true type font to force font smoothing of any size.
  Copyright (c) 1997 by Wei Ke (AKA KEWEI)

  Author: Wei Ke

  Usage: ttfgasp fontfile.ttf
  output: fontfile.ttf - "gasp" added
          fontfile.ttf.$xx - backup of the original

  To Compile: DCC32.EXE -CC %1
}  

program TTFgasp;

uses SysUtils;

type
  TByteArray = array [0..MaxInt-1] of Byte;
  PByte = ^TByteArray;

  TLongArray = array [0..(MaxInt div SizeOf(Longint))-1] of Longint;
  PLong = ^TLongArray;

  {$A-}
  TTableDir = record
    tag: packed array [1..4] of char;
    checksum: Longint;
    offset: Longint;
    len: Longint
  end;
  PTableDir = ^TTableDir;

  TFontHead = record
    version: Longint;
    numTables: Word;
    searchRange: Word;
    entrySelector: Word;
    rangeShift: Word
  end;
  {$A+}

  TTable = record
    pdir: PTableDir;
    pmem: PByte
  end;

const
  MAXTABLE = 256;

var
  head: TFontHead;
  vtable: array [0..MAXTABLE] of TTable;
  nTables: Longint;

function SwapLong(l: Longint): Longint; assembler;
asm
  mov   eax, l
  bswap eax
end;

function SwapShort(w: Word): Word; assembler;
asm
  movzx eax, w
  xchg  ah, al
end;

procedure ReadFontHead(var f: File);
begin
  Seek(f, 0);
  BlockRead(f, head, SizeOf(head))
end;

procedure ReadTableDirs(var f: File);
var
  i: Integer;
begin
  nTables := 0;
  Seek(f, SizeOf(TFontHead));
  for i := 1 to SwapShort(head.numTables) do begin
    New(vtable[nTables].pdir);
    BlockRead(f, vtable[nTables].pdir^, SizeOf(TTableDir));
    Inc(nTables)
  end
end;

procedure ReadTables(var f: File);
var
  i, j, l: Integer;
begin
  for i := 0 to nTables-1 do begin
    l := SwapLong(vtable[i].pdir^.len);
    GetMem(vtable[i].pmem, (l+3) and (not 3));
    for j := l to ((l+3) and (not 3))-1 do
      vtable[i].pmem^[j] := 0;
    Seek(f, SwapLong(vtable[i].pdir^.offset));
    BlockRead(f, vtable[i].pmem^, l)
  end
end;

procedure AddGASPTable;
var
  i: Integer;
begin
  i := 0;
  while (i < nTables) and (vtable[i].pdir^.tag <> 'gasp') do
    Inc(i);

  if i >= nTables then begin
    New(vtable[nTables].pdir);
    Inc(nTables)
  end;

  vtable[i].pdir^.tag := 'gasp';
  vtable[i].pdir^.len := SwapLong(8);
  GetMem(vtable[i].pmem, 8);
  vtable[i].pmem^[0] := 0;
  vtable[i].pmem^[1] := 0;

  vtable[i].pmem^[2] := 0;
  vtable[i].pmem^[3] := 1;

  vtable[i].pmem^[4] := $FF;
  vtable[i].pmem^[5] := $FF;

  vtable[i].pmem^[6] := $00;
  vtable[i].pmem^[7] := $03
end;

procedure CalcTableChecksum;
var
  s: Longint;
  p: PLong;
  i, j: Integer;
begin
  for j := 0 to nTables-1 do begin
    s := 0;
    p := PLong(vtable[j].pmem);
    for i := 0 to ((SwapLong(vtable[j].pdir^.len)+3) div SizeOf(Longint))-1 do begin
      Inc(s, SwapLong(p^[i]))
    end;
    vtable[j].pdir^.checksum := SwapLong(s)
  end
end;

procedure SortTables;
var
  i, j: Integer;
  temp: TTable;
begin
  for i := 0 to nTables-2 do
    for j := i+1 to nTables-1 do
      if vtable[i].pdir^.tag > vtable[j].pdir^.tag then begin
        temp := vtable[i];
        vtable[i] := vtable[j];
        vtable[j] := temp
      end
end;

procedure WriteTables(var f: File);
var
  i: Integer;
begin
  Seek(f, SizeOf(TFontHead)+nTables*SizeOf(TTableDir));
  for i := 0 to nTables-1 do begin
    vtable[i].pdir.offset := SwapLong(FilePos(f));
    BlockWrite(f, vtable[i].pmem^, (SwapLong(vtable[i].pdir^.len)+3) and (not 3))
  end
end;

procedure WriteTableDirs(var f: File);
var
  i: Integer;
begin
  Seek(f, SizeOf(TFontHead));
  for i := 0 to nTables-1 do begin
    BlockWrite(f, vtable[i].pdir^, SizeOf(TTableDir))
  end
end;

procedure WriteFontHead(var f: File);
var
  p, n: Integer;
begin
  head.numTables := SwapShort(Word(nTables));
  p := 1;
  n := 0;
  while p*2 <= nTables do begin
    p := p*2;
    Inc(n)
  end;

  head.searchRange := SwapShort(Word(p*16));
  head.entrySelector := SwapShort(Word(n));
  head.rangeShift := SwapShort(Word(nTables*16-p));

  Seek(f, 0);
  BlockWrite(f, head, SizeOf(head))
end;

function RenameToBak(s: String): String;
var
  f: File;
  i: Integer;
begin
  i := 0;
  while (i < 256) and FileExists(ExpandFileName(s)+'.$'+IntToHex(i, 2)) do
    Inc(i);

  if i >= 256 then begin
    result := ExpandFileName(s)+'.$$$';
    if FileExists(result) then begin
      AssignFile(f, result);
      Erase(f)
    end
  end else
    result := ExpandFileName(s)+'.$'+IntToHex(i, 2);

  AssignFile(f, s);
  Rename(f, result)
end;

var
  fin, fout: File;

begin
  AssignFile(fin, RenameToBak(ParamStr(1)));
  AssignFile(fout, ParamStr(1));
  Reset(fin, 1);
  ReadFontHead(fin);
  ReadTableDirs(fin);
  ReadTables(fin);
  Close(fin);
  AddGASPTable;
  SortTables;
  CalcTableChecksum;
  Rewrite(fout, 1);
  WriteTables(fout);
  WriteTableDirs(fout);
  WriteFontHead(fout);
  Close(fout)
end.

